import logging
import math
from pvpower.estimator import Estimator, SVMEstimator, FullVectorizer, CoreVectorizer, SunshineVectorizer, PlusVisibilityVectorizer, SushinePlusCloudCoverVectorizer, PlusSunshineVectorizer, PlusCloudCoverVectorizer, PlusVisibilitySunshineVectorizer, PlusVisibilityCloudCoverVectorizer, PlusVisibilityFogCloudCoverVectorizer
from pvpower.traindata import TrainData


class TrainRun:

    def __init__(self, estimator: Estimator, train_data: TrainData):
        example_data, validation_data = train_data.split()
        self.validation_samples  = [record for record in validation_data.samples]
        self.estimator = estimator
        self.estimator.retrain(example_data)
        self.predictions = [estimator.predict(validation_record) for validation_record in self.validation_samples]
        self.estimator.retrain(train_data)  # retrain with all data

    def __score(self, real, predicted):
        rounded_real = round(real/10)*10
        rounded_predicted = round(predicted/10)*10
        score = abs(rounded_real - rounded_predicted)
        return score

    @property
    def score(self) -> float:
        scores = []
        for i in range(len(self.validation_samples)):
            scores.append(self.__score(self.validation_samples[i].power_watt, self.predictions[i]))

        scores = sorted(scores)
        ignore_size = int(len(scores) * 0.1)
        if ignore_size <= 0:
            ignore_size = 1
        scores = scores[ignore_size:-ignore_size]
        if len(scores) == 0:
            return 999999
        else:
            return round(sum(scores) / len(scores), 2)

    def __lt__(self, other):
        return self.score < other.score

    def __eq__(self, other):
        return self.score == other.score

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        max_lines = 1000
        txt = str(self.estimator) + "\n"
        txt += "score:  " + str(self.score) + " (smaller is better)\n"
        txt += '{:25s}   {:10s} {:10s}    {:10s}          {:14s} {:14s} {:14s} {:14s} {:14s}         \n'.format("time", "real", "predicted", "score", "irradiance", "sunshine", "cloud_cover", "visibility", "proba.fog")
        for i in range(0, len(self.validation_samples)):
            if i > max_lines:
                txt += '....'
                break
            txt += '{:<25s}   {:<10d} {:<10d}    {:<10s}          {:<14d} {:<14d} {:<14s} {:<14d}  {:<14d}        \n'.format(self.validation_samples[i].time.strftime("%d.%b %H:%M") + " (" + self.validation_samples[i].time_utc.strftime("%H:%M") + " utc)",
                                                                                                                               self.validation_samples[i].power_watt,
                                                                                                                               self.predictions[i],
                                                                                                                               "[" + str(self.__score(self.validation_samples[i].power_watt, self.predictions[i])) + "]",
                                                                                                                               self.validation_samples[i].irradiance,
                                                                                                                               self.validation_samples[i].sunshine,
                                                                                                                               str(self.validation_samples[i].cloud_cover_effective) + "%",
                                                                                                                               self.validation_samples[i].visibility,
                                                                                                                               self.validation_samples[i].probability_for_fog)
        return txt



class TrainingCenter:

    def new_estimator(self, trainData: TrainData) -> Estimator:
        estimators = [SVMEstimator(CoreVectorizer()),
                      SVMEstimator(FullVectorizer()),
                      SVMEstimator(SunshineVectorizer()),
                      SVMEstimator(SushinePlusCloudCoverVectorizer()),
                      SVMEstimator(PlusVisibilityVectorizer()),
                      SVMEstimator(PlusSunshineVectorizer()),
                      SVMEstimator(PlusCloudCoverVectorizer()),
                      SVMEstimator(PlusVisibilitySunshineVectorizer()),
                      SVMEstimator(PlusVisibilityCloudCoverVectorizer()),
                      SVMEstimator(PlusVisibilityFogCloudCoverVectorizer())]

        logging.info("train estimators with " + str(len(trainData.samples)) + " samples")
        rounds = 6
        train_runs = []
        for estimator in estimators:
            runs = sorted([TrainRun(estimator, trainData.rotated(round(i * 100 / rounds))) for i in range(0, rounds)])
            cleaned_runs = [run for run in runs if run.score < 10000]
            if len(cleaned_runs) > 0:
                runs = cleaned_runs
            train_runs.append(runs[int(len(runs)*0.5)])

        train_runs = sorted(train_runs, reverse=True)
        for run in train_runs:
            logging.debug(run)
        best_train = train_runs[-1]
        logging.info("new estimator trained \n" + str(best_train))
        return best_train.estimator



import logging
import pickle
from os import path, makedirs
from os.path import exists
from appdirs import user_cache_dir
from threading import Thread, Lock
from datetime import datetime, timedelta
from pvpower.weather_forecast import WeatherForecast
from pvpower.traindata import TrainSampleLog, TrainData
from pvpower.estimator import Estimator, DelegatingEstimator, SVMEstimator, FullVectorizer
from pvpower.trainingcenter import TrainingCenter


class PersistentEstimator(DelegatingEstimator):

    def __init__(self, train_log: TrainSampleLog):
        self.__train_log = train_log
        self.__lock = Lock()
        self.__training_center = TrainingCenter()
        self.__date_last_retrain_initiated = datetime.fromtimestamp(0)
        super(PersistentEstimator, self).__init__(PersistentEstimator.__load())

    def predict(self, sample: WeatherForecast) -> int:
        try:
            return super(PersistentEstimator, self).predict(sample)
        finally:
            try:
                retrain_period_days = 1 + (self.duration_sec_last_train() * 7 * 24)  # min 1 day + 7 days per 1 sec traintime
                with self.__lock:
                    if datetime.now() > (self.__date_last_retrain_initiated + timedelta(hours=int(retrain_period_days*24))):
                        self.__date_last_retrain_initiated = datetime.now()
                        Thread(target=self.retrain, args=(self.__train_log.all(),), daemon=True).start()
            except Exception as e:
                logging.warning("error occurred checking last train duration of estimator", e)
                Thread(target=self.retrain, args=(self.__train_log.all(),), daemon=True).start()

    def retrain(self, train_data: TrainData):
        self._estimator = self.__training_center.new_estimator(train_data)
        self.__store(self._estimator)

    @staticmethod
    def __store(estimator: Estimator):
        try:
            with open(PersistentEstimator.__filename(), 'wb') as file:
                pickle.dump(estimator, file)
        except Exception as e:
            logging.warning("error occurred storing pickled estimator " + str(e))

    @staticmethod
    def __load() -> Estimator:
        estimator = SVMEstimator(FullVectorizer())
        if exists(PersistentEstimator.__filename()):
            try:
                with open(PersistentEstimator.__filename(), 'rb') as file:
                    estimator = pickle.load(file)
                    logging.debug("estimator " + str(estimator) + " loaded from pickle file (" + PersistentEstimator.__filename() + ")")
            except Exception as e:
                logging.warning("error occurred loading estimator " + str(e))
        return estimator

    @staticmethod
    def __filename():
        cache_dir = user_cache_dir("pv_forecast", appauthor=False)
        if not path.exists(cache_dir):
            makedirs(cache_dir)
        return path.join(cache_dir, 'Estimator.p')

    def __str__(self):
        return str(self._estimator)




import logging
import os
import pytz
import shutil
import gzip
import tempfile
from time import sleep
from appdirs import site_data_dir
from threading import Thread
from pathlib import Path
from os.path import exists
from threading import RLock
from datetime import datetime, timedelta
from typing import List, Optional
from pvpower.weather_forecast import WeatherForecast



class LabelledWeatherForecast(WeatherForecast):

    def __init__(self,
                 time: datetime,
                 irradiance: int,
                 sunshine: int,
                 cloud_cover_effective: int,
                 probability_for_fog: int,
                 visibility: int,
                 power_watt: int):
        super().__init__(time, irradiance, sunshine, cloud_cover_effective, probability_for_fog, visibility)
        self.power_watt = power_watt


    @staticmethod
    def create(weather_forecast: WeatherForecast, power_watt: int, time: datetime = None):

        return LabelledWeatherForecast(weather_forecast.time_utc if time is None else time,
                                       weather_forecast.irradiance,
                                       weather_forecast.sunshine,
                                       weather_forecast.cloud_cover_effective,
                                       weather_forecast.probability_for_fog,
                                       weather_forecast.visibility,
                                       power_watt)

    @staticmethod
    def __to_string(value):
        return "" if value is None else str(value)

    @staticmethod
    def csv_header() -> str:
        return "utc_time;real_pv_power;irradiance;sunshine;cloud_cover;probability_for_fog;visibility"

    @staticmethod
    def __to_int(txt) -> Optional[int]:
        if len(txt) > 0:
            return int(float(txt))
        else:
            return None

    def to_csv(self) -> str:
        utc_time = self.time_utc
        return utc_time.isoformat() + ";" + \
               LabelledWeatherForecast.__to_string(self.power_watt) + ";" + \
               LabelledWeatherForecast.__to_string(self.irradiance) + ";" + \
               LabelledWeatherForecast.__to_string(self.sunshine) + ";" + \
               LabelledWeatherForecast.__to_string(self.cloud_cover_effective) + ";" + \
               LabelledWeatherForecast.__to_string(self.probability_for_fog) + ";" + \
               LabelledWeatherForecast.__to_string(self.visibility)

    @staticmethod
    def from_csv(line: str):
        parts = line.split(";")
        utc_time = datetime.fromisoformat(parts[0])
        real_pv_power = LabelledWeatherForecast.__to_int(parts[1])
        irradiance = LabelledWeatherForecast.__to_int(parts[2])
        sunshine = LabelledWeatherForecast.__to_int(parts[3])
        cloud_cover_effective = LabelledWeatherForecast.__to_int(parts[4])
        probability_for_fog = LabelledWeatherForecast.__to_int(parts[5])
        visibility = LabelledWeatherForecast.__to_int(parts[6])
        sample = LabelledWeatherForecast(utc_time, irradiance, sunshine, cloud_cover_effective, probability_for_fog, visibility, real_pv_power)
        return sample


    def __str__(self):
        return super().__str__() + ", power_watt=" + str(self.power_watt)


class TrainData:

    def __init__(self, samples: List[LabelledWeatherForecast]):
        self.samples = self.__clean_data(samples)

    def __iter__(self):
        return iter(self.samples)

    def __len__(self):
        return len(self.samples)

    @staticmethod
    def __clean_data(samples: List[LabelledWeatherForecast]) -> List[LabelledWeatherForecast]:
        seen = list()
        return list(filter(lambda sample: seen.append(sample.time_utc) is None if sample.time_utc not in seen else False, samples))  # remove duplicates

    def rotated(self, offset_percent: int):
        step = int(offset_percent * len(self.samples) / 100)
        return TrainData(self.samples[step:] + self.samples[:step])

    def split(self, ratio_percent:int = 67):
        size = len(self.samples)
        left = []
        right = []
        for i in range(0, size):
            if i > round(size * ratio_percent / 100, 0):
                right.append(self.samples[i])
            else:
                left.append(self.samples[i])
        return TrainData(left), TrainData(right)

    def __str__(self):
        return "train data size " + str(len(self.samples)) + " (" + self.samples[0].time.strftime("%Y.%m.%dT%H:%M") + " ... " + self.samples[-1].time.strftime("%Y.%m.%dT%H:%M") + ")"


class TrainSampleLog:
    COMPACTION_PERIOD_DAYS = 15
    FILENAME = "train.csv.gz"

    def __init__(self, dirname: str = None):
        self.lock = RLock()
        self.__dirname = dirname if dirname is not None else site_data_dir("pv_power", appauthor=False)
        self.__last_compaction_time = datetime.now() - timedelta(days=self.COMPACTION_PERIOD_DAYS*2)

    def filename(self):
        fn = os.path.join(self.__dirname, self.FILENAME)
        if not exists(fn):
            directory = Path(fn).parent
            if not exists(directory):
                os.makedirs(directory)
        return fn

    def append(self, sample: LabelledWeatherForecast):
        with self.lock:
            compr_fn = self.filename()
            with gzip.open(compr_fn, "ab") as file:
                line = sample.to_csv() + "\n"
                file.write(line.encode(encoding='UTF-8'))

        if datetime.now() > (self.__last_compaction_time + timedelta(days=self.COMPACTION_PERIOD_DAYS)):
            self.__last_compaction_time = datetime.now()
            Thread(target=self.compact, args=(15,), daemon=True).start()

    def all(self) -> TrainData:
        with self.lock:
            compr_fn = self.filename()
            if exists(compr_fn):
                try:
                    with gzip.open(compr_fn, "rb") as file:
                        lines = [raw_line.decode('UTF-8').strip() for raw_line in file.readlines()]
                        samples = []
                        for line in lines:
                            try:
                                samples.append(LabelledWeatherForecast.from_csv(line))
                            except Exception as e:
                                pass
                        return TrainData(samples)
                except Exception as e:
                    logging.warning("error occurred loading " + compr_fn + " " + str(e))
        return TrainData([])

    def compact(self, delay_sec:int = 0):
        sleep(delay_sec)

        fn = self.filename()
        train_data = self.all()
        min_datetime = datetime.now() - timedelta(days=4*365)

        with tempfile.TemporaryDirectory() as tmp_dir:
            temp_file = os.path.join(tmp_dir, 'traindata.csv')
            with gzip.open(temp_file, "wb") as file:
                file.write((LabelledWeatherForecast.csv_header() + "\n").encode(encoding='UTF-8'))
                num_survived = 0
                for sample in train_data:
                    expired = sample.time_utc < min_datetime.astimezone(pytz.UTC)
                    if not expired:
                        num_survived += 1
                        line = sample.to_csv() + "\n"
                        file.write(line.encode(encoding='UTF-8'))
            shutil.move(temp_file, fn)
            logging.info("train file " + fn + " compacted")

    def __str__(self):
        return "\n".join([sample.to_csv() for sample in self.all()])






import logging
from datetime import datetime, timedelta
from typing import Optional
from pvpower.weather_forecast import WeatherStation, WeatherForecast
from pvpower.traindata import LabelledWeatherForecast, TrainSampleLog
from pvpower.estimator import Estimator
from pvpower.persistent_estimator import PersistentEstimator


class ValueRecorder:

    def __init__(self):
        self.start_time = datetime.strptime(datetime.now().strftime("%d.%m.%Y %H") + ":00", "%d.%m.%Y %H:%M")
        self.end_time = self.start_time + timedelta(minutes=60)
        self.__power_values = []
        #logging.debug("value recorder created (" + str(self) + ")")

    def empty(self) -> bool:
        return len(self.__power_values) == 0

    def is_expired(self):
        return datetime.now() >= self.end_time

    def add(self, value: int):
        self.__power_values.append(value)
        #logging.debug("record added to value recorder (" + str(self) + ")")

    @property
    def average(self) -> Optional[int]:
        if len(self.__power_values) == 0:
            return None
        else:
            return int(round(sum(self.__power_values) / len(self.__power_values), 0))

    def __str__(self):
        return self.start_time.strftime("%Y.%m.%d %H:%M") + " -> " + self.end_time.strftime("%Y.%m.%d %H:%M") + "  average power: " + str(self.average) + " num probes: " + str(len(self.__power_values))


class PvPowerForecast:

    def __init__(self, station_id: str, pv_forecast_dir: None, estimator: Estimator = None):
        self.__train_value_recorder = ValueRecorder()
        self.weather_forecast_service = WeatherStation(station_id)
        self.train_log = TrainSampleLog(pv_forecast_dir)
        self.__estimator = estimator if estimator is not None else PersistentEstimator(self.train_log)

    def add_current_power_reading(self, real_power: int):
        if self.__train_value_recorder.is_expired():
            try:
                if not self.__train_value_recorder.empty():
                    weather_sample = self.weather_forecast_service.forecast(self.__train_value_recorder.start_time)
                    if weather_sample is not None:
                        annotated_sample = LabelledWeatherForecast.create(weather_sample,
                                                                          self.__train_value_recorder.average,
                                                                          time=self.__train_value_recorder.start_time)
                        logging.info("add train sample " + str(annotated_sample))
                        self.train_log.append(annotated_sample)
            finally:
                self.__train_value_recorder = ValueRecorder()
        self.__train_value_recorder.add(real_power)

    def predict(self, time: datetime) -> Optional[int]:
        sample = self.weather_forecast_service.forecast(time)
        if sample is None:
            logging.info("could not predict power. Reason: no weather forecast data available (requested date time: " + time.strftime("%Y.%m.%d %H:%M") +
                         ". Available date time range: " + self.weather_forecast_service.forcast_from().strftime("%Y.%m.%d %H:%M") +
                         " -> " + self.weather_forecast_service.forcast_to().strftime("%Y.%m.%d %H:%M")  + "). Returning None")
            return None
        else:
            return self.predict_by_weather_forecast(sample)

    def predict_by_weather_forecast(self, sample: WeatherForecast) -> int:
        return self.__estimator.predict(sample)

    def __str__(self):
        return str(self.__estimator)
